from __future__ import unicode_literals


def get_cardinal_category(num_string, lang):
    n, i, v, w, f, t = get_parts_of_num(num_string)

    if lang in LOCALE_FUNCTIONS:
        return LOCALE_FUNCTIONS[lang](n, i, v, w, f, t)
    else:
        raise Exception('Unknown locale code.')


def get_parts_of_num(num_string):
    '''
    Gets the different parts of a number in order to calculate which plurality
    rule to apply.

    Parts are specified at this URL:
    http://unicode.org/reports/tr35/tr35-numbers.html#Operands

    :returns:
        n: absolute value of the source number (integer and decimals).
        i: integer digits of n.
        v: number of visible fraction digits in n, with trailing zeros.
        w: number of visible fraction digits in n, without trailing zeros.
        f: visible fractional digits in n, with trailing zeros.
        t: visible fractional digits in n, without trailing zeros.
    '''
    n = num_string

    decimal_split = n.split('.')
    i = int(decimal_split[0])

    if len(decimal_split) != 2:
        return n, i, 0, 0, 0, 0

    decimal_part = decimal_split[1]
    v = len(decimal_part)
    f = int(decimal_part)

    dec_part_no_trailing_zeros = decimal_part.rstrip('0')
    if not dec_part_no_trailing_zeros:
        return n, i, v, 0, f, 0

    w = len(dec_part_no_trailing_zeros)
    t = int(dec_part_no_trailing_zeros)

    return n, i, v, w, f, t


def brazilian_pt(n, i, v, w, f, t):
    if 0 <= i < 2 and f == 0:
        return 'one'

    return 'other'


def czech(n, i, v, w, f, t):
    if i == 1 and v == 0:
        return 'one'

    if 2 <= i <= 4 and v == 0:
        return 'few'

    if v != 0:
        return 'many'

    return 'other'


def french(n, i, v, w, f, t):
    if i == 0 or i == 1:
        return 'one'

    return 'other'


def one_or_other(n, i, v, w, f, t):
    if i == 1 and v == 0:
        return 'one'

    return 'other'


def other(n, i, v, w, f, t):
    return 'other'


LOCALE_FUNCTIONS = {
    'cs': czech,
    'de': one_or_other,
    'en': one_or_other,
    'fr': french,
    'it': one_or_other,
    'ja': other,
    'nl': one_or_other,
    'pt': brazilian_pt,
    'pt_PT': one_or_other,
    'zh': other
}



from __future__ import unicode_literals

from parsimonious.grammar import Grammar

ICUMessageFormat = Grammar('''
    message_format_pattern =
        (message_format_element/octothorpe/string)*

    message_format_element = "{" _ id ("," element_format)? _ "}"

    element_format =
        (_ replace_type _ "," _ plural_format_pattern _) /
        (_ replace_type _ "," _ select_format_pattern _) /
        (_ id arg_style_pattern*)

    plural_format_pattern = offset? (plural_form)+
    plural_form = _ plural_key _ "{" _ message_format_pattern _ "}"
    plural_key = id / ("=" decimal)

    select_format_pattern = select_form+
    select_form = _ id _ "{" _ message_format_pattern _ "}"

    arg_style_pattern = _ "," _ id _

    offset = _ "offset" _ ":" _ digits _

    octothorpe   = ~"#"
    string       = (~"\w+"/~"[^{}]+"/_)+
    id           = ~"\w+"i
    replace_type = ~"\w+"i
    decimal      = ~"[0-9]+(\.[0-9]+)?"
    digits       = ~"[0-9]+"
    _            = ~"\s*"
    ''')



from __future__ import unicode_literals

from builtins import str
from decimal import Decimal
from parsimonious.nodes import NodeVisitor
from toolz import merge

from pyseeyou.locales import get_cardinal_category


class ICUNodeVisitor(NodeVisitor):
    def __init__(self, options, lang='en'):
        '''
        ICUNodeVisitor is a walker for the generated parse tree from
        pyseeyou.grammar.ICUMessageFormat.

        Calling visit() on an instance of the ICUNodeVisitor will traverse the
        tree, returning a fully formed string, using the self.options dict as
        the values to replace any IDs that were parsed using the grammar.

        :param options: Values used to replace parsed IDs with.
        :param lang: Language used to derive pluralisation phrase of number
        '''
        self.options = options
        self.lang = lang

    def generic_visit(self, node, visited_children):
        '''
        Each of the methods below beginning with visit_ are called when
        visiting a node of that type. These methods deal with how to parse and
        process the data that has reached that node.

        This generic_visit method deals with visiting nodes that have no
        specific type.
        '''
        if visited_children:
            visited_children = self._filter_none(visited_children)

        if len(visited_children) > 1:
            return merge(visited_children)

        elif len(visited_children) == 1:
            return visited_children[0]

    def visit_message_format_pattern(self, node, visited_children):
        text = ''

        for item in visited_children:
            if not item:
                continue

            if isinstance(item, str):
                text += item

            else:
                for key in item:
                    text += self._get_formed_string(item, key)

        return text

    def visit_message_format_element(self, node, visited_children):
        visited_children = self._filter_none(visited_children)
        key = visited_children[0].get('key')

        if len(visited_children) == 1:
            return {key: None}
        else:
            values = visited_children[1]
            return {key: values}

    def visit_element_format(self, node, visited_children):
        return visited_children[0]

    def visit_select_format_pattern(self, node, visited_children):
        return merge(visited_children)

    def visit_select_form(self, node, visited_children):
        return self._get_key_value(visited_children)

    def visit_plural_format_pattern(self, node, visited_children):
        visited_children = self._filter_none(visited_children)
        return merge(visited_children)

    def visit_plural_form(self, node, visited_children):
        return self._get_key_value(visited_children)

    def visit_plural_key(self, node, visited_children):
        if isinstance(visited_children[0], str):
            return {'key': visited_children[0]}

        return visited_children[0]

    def visit_arg_style_pattern(self, node, visited_children):
        pass

    def visit_offset(self, node, visited_children):
        for child in visited_children:
            if isinstance(child, int):
                return {'offset': child}

    def visit_octothorpe(self, node, visited_children):
        return str(node.text)

    def visit_string(self, node, visited_children):
        return str(node.text)

    def visit_id(self, node, visited_children):
        return {'key': str(node.text)}

    def visit_replace_type(self, node, visited_children):
        return {'replace_type': str(node.text)}

    def visit_decimal(self, node, visited_children):
        return str(node.text)

    def visit_digits(self, node, visited_children):
        return int(node.text)

    def visit__(self, node, visited_children):
        pass

    def _get_formed_string(self, item, key):
        # Direct replacement, {} style
        if not item[key]:
            return self.options[key]

        replace_type = item[key]['replace_type']
        if replace_type.lower() == 'select':
            return self._select_replace(item[key], key)

        elif replace_type.lower() == 'plural':
            return self._plural_replace(item[key], key)

    def _select_replace(self, item, key):
        if key in self.options:
            return item[self.options[key]]

        else:
            return item['other']

    def _plural_replace(self, item, key):
        str_key = str(self.options[key])

        if 'offset' in item:
            dec_str_key = Decimal(str_key)
            dec_str_key -= item['offset']

            if dec_str_key < 0:
                str_key = '0'
            else:
                str_key = str(dec_str_key)

        if str_key in item:
            return item[str_key]

        else:
            plural_key = get_cardinal_category(str_key, self.lang)

            if '#' in item[plural_key]:
                return item[plural_key].replace(
                    '#', str(str_key))

            else:
                return item[plural_key]

    def _get_key_value(self, items):
        key, value = None, None
        for item in items:
            if not item:
                continue

            elif isinstance(item, str):
                value = item

            elif isinstance(item, dict):
                key = item['key']

        return {key: value}

    def _filter_none(self, items):
        return [item for item in items if item is not None]



import pytest

from pyseeyou.locales import (
    get_parts_of_num, brazilian_pt, czech, french, one_or_other, other)


def test_get_parts_of_num():
    assert get_parts_of_num('1') == ('1', 1, 0, 0, 0, 0)
    assert get_parts_of_num('1.0') == ('1.0', 1, 1, 0, 0, 0)
    assert get_parts_of_num('1.00') == ('1.00', 1, 2, 0, 0, 0)
    assert get_parts_of_num('1.3') == ('1.3', 1, 1, 1, 3, 3)
    assert get_parts_of_num('1.30') == ('1.30', 1, 2, 1, 30, 3)
    assert get_parts_of_num('1.03') == ('1.03', 1, 2, 2, 3, 3)
    assert get_parts_of_num('1.230') == ('1.230', 1, 3, 2, 230, 23)


def test_brazilian_pt():
    assert brazilian_pt(*get_parts_of_num('0')) == 'one'
    assert brazilian_pt(*get_parts_of_num('1')) == 'one'
    assert brazilian_pt(*get_parts_of_num('1.0')) == 'one'
    assert brazilian_pt(*get_parts_of_num('1.3')) == 'other'
    assert brazilian_pt(*get_parts_of_num('2')) == 'other'
    assert brazilian_pt(*get_parts_of_num('4')) == 'other'
    assert brazilian_pt(*get_parts_of_num('9')) == 'other'
    assert brazilian_pt(*get_parts_of_num('13')) == 'other'


def test_czech():
    assert czech(*get_parts_of_num('0')) == 'other'
    assert czech(*get_parts_of_num('1')) == 'one'
    assert czech(*get_parts_of_num('1.0')) == 'many'
    assert czech(*get_parts_of_num('1.3')) == 'many'
    assert czech(*get_parts_of_num('2')) == 'few'
    assert czech(*get_parts_of_num('2.5')) == 'many'
    assert czech(*get_parts_of_num('4')) == 'few'
    assert czech(*get_parts_of_num('13')) == 'other'
    assert czech(*get_parts_of_num('13.5')) == 'many'


def test_french():
    assert french(*get_parts_of_num('0')) == 'one'
    assert french(*get_parts_of_num('1')) == 'one'
    assert french(*get_parts_of_num('1.0')) == 'one'
    assert french(*get_parts_of_num('1.3')) == 'one'
    assert french(*get_parts_of_num('2')) == 'other'
    assert french(*get_parts_of_num('4')) == 'other'
    assert french(*get_parts_of_num('9')) == 'other'
    assert french(*get_parts_of_num('13')) == 'other'


def test_other():
    assert other(*get_parts_of_num('0')) == 'other'
    assert other(*get_parts_of_num('1')) == 'other'
    assert other(*get_parts_of_num('1.0')) == 'other'
    assert other(*get_parts_of_num('1.3')) == 'other'
    assert other(*get_parts_of_num('2')) == 'other'
    assert other(*get_parts_of_num('4')) == 'other'
    assert other(*get_parts_of_num('9')) == 'other'
    assert other(*get_parts_of_num('13')) == 'other'


def test_one_or_other():
    assert one_or_other(*get_parts_of_num('0')) == 'other'
    assert one_or_other(*get_parts_of_num('1')) == 'one'
    assert one_or_other(*get_parts_of_num('1.0')) == 'other'
    assert one_or_other(*get_parts_of_num('1.3')) == 'other'
    assert one_or_other(*get_parts_of_num('2')) == 'other'
    assert one_or_other(*get_parts_of_num('4')) == 'other'
    assert one_or_other(*get_parts_of_num('9')) == 'other'
    assert one_or_other(*get_parts_of_num('13')) == 'other'
